# granola_client/__init__.py
__version__ = "0.2.0"

from .client import GranolaClient
from .errors import (
    GranolaAPIError, GranolaAuthError, GranolaRateLimitError,
    GranolaTimeoutError, GranolaValidationError
)
from .types import (
    # Client/HTTP Options
    ClientOpts, HttpOpts,
    # Core Models
    Document, DocumentsResponse,
    DocumentMetadata, TranscriptSegment, PanelTemplate,
    # Entity Models
    Person,
    # Feature/Integration Models
    FeatureFlagsResponse,
    # Subscription Models
    SubscriptionsResponse,
    # Payload Models
    UpdateDocumentPayload, UpdateDocumentPanelPayload,
    # Filter Models
    GetDocumentsFilters,
    # Add other important Pydantic models you want to be easily accessible
)
from .pagination import PaginatedResponse # This is also a Pydantic model now

# For easier imports like: from granola_client import GranolaClient
__all__ = [
    "GranolaClient",
    "GranolaAPIError",
    "GranolaAuthError",
    "GranolaRateLimitError",
    "GranolaTimeoutError",
    "GranolaValidationError",
    # Options
    "ClientOpts", "HttpOpts",
    # Models
    "Document", "DocumentsResponse",
    "DocumentMetadata", "TranscriptSegment", "PanelTemplate", "Person",
    "FeatureFlagsResponse", "SubscriptionsResponse",
    "UpdateDocumentPayload", "UpdateDocumentPanelPayload", "GetDocumentsFilters",
    "PaginatedResponse",
    "__version__",
]

# Configure basic logging if the library user hasn't
import logging
logging.getLogger(__name__).addHandler(logging.NullHandler())



import asyncio
import httpx
import platform
import json # For logging bodies, Pydantic handles main ser/de
from typing import Optional, Callable, Awaitable, TypeVar, Dict, Any, Union, Type, cast, List
import logging

from pydantic import BaseModel, ValidationError, HttpUrl

from .types import HttpOpts # HttpOpts is now a Pydantic model
from .errors import (
    GranolaAPIError, GranolaAuthError, GranolaRateLimitError,
    GranolaTimeoutError, GranolaValidationError
)

logger = logging.getLogger(__name__)

T = TypeVar('T', bound=BaseModel) # Response model type
P = TypeVar('P', bound=BaseModel) # Payload model type

class HttpClient:
    def __init__(
        self,
        token: Optional[str] = None,
        base_url: Union[str, HttpUrl] = "https://api.granola.ai", # Default from ClientOpts
        opts: Optional[HttpOpts] = None, # HttpOpts is Pydantic model
    ):
        # If opts is None, create a default HttpOpts instance
        # If opts is a dict, Pydantic will parse it
        parsed_opts = HttpOpts.model_validate(opts if opts is not None else {})

        self._token: Optional[str] = token
        self.base_url_str: str = str(base_url).rstrip("/") # base_url from ClientOpts is HttpUrl
        self.timeout_ms: int = cast(int, parsed_opts.timeout)
        self.retries: int = cast(int, parsed_opts.retries)

        self.app_version: str = cast(str, parsed_opts.app_version)
        self.client_type: str = cast(str, parsed_opts.client_type)

        self.client_platform: str = parsed_opts.client_platform or platform.system().lower()
        self.client_architecture: str = parsed_opts.client_architecture or platform.machine().lower()
        self.electron_version: str = cast(str, parsed_opts.electron_version)
        self.chrome_version: str = cast(str, parsed_opts.chrome_version)
        self.node_version: str = cast(str, parsed_opts.node_version)

        os_ver = platform.release()
        if self.client_platform == "darwin":
            os_ver = platform.mac_ver()[0]
        elif self.client_platform == "win32": # platform.system() returns 'Windows'
            self.client_platform = "windows" # align with common usage
            os_ver = platform.win32_ver()[0]

        self.os_version: str = parsed_opts.os_version or os_ver
        self.os_build: str = parsed_opts.os_build or "" # Still hard to get reliably

        self.client_headers: Dict[str, str] = cast(Dict[str,str], parsed_opts.client_headers)

        self._token_provider: Optional[Callable[[], Awaitable[str]]] = None
        self._token_fetch_lock = asyncio.Lock()

        self._client: httpx.AsyncClient = httpx.AsyncClient(timeout=self.timeout_ms / 1000.0)

    def set_token(self, token: str) -> None:
        self._token = token

    def set_token_provider(self, provider: Callable[[], Awaitable[str]]) -> None:
        self._token_provider = provider

    async def _ensure_token(self) -> None:
        if self._token:
            return
        if not self._token_provider:
            logger.warning("No API token set and no token provider configured.")
            return

        async with self._token_fetch_lock:
            if self._token: return
            try:
                logger.info("Fetching API token using provider.")
                self._token = await self._token_provider()
                if not self._token:
                    raise GranolaAuthError("Token provider returned an empty token.")
                logger.info("API token successfully fetched and set.")
            except Exception as e:
                logger.error(f"Error fetching token via provider: {e}", exc_info=True)
                raise GranolaAuthError(f"Failed to retrieve authentication token via provider: {e}") from e

    def _get_backoff_delay(self, attempt: int, retry_after_header: Optional[str] = None) -> float:
        if retry_after_header:
            try: return float(retry_after_header)
            except ValueError: pass
        return 0.25 * (2 ** attempt)

    async def _request_raw(
        self, method: str, path: str, body_data: Optional[Union[Dict[str, Any], List[Any], str]] = None
    ) -> httpx.Response:
        await self._ensure_token()
        url = f"{self.base_url_str}{path}"

        headers: Dict[str, str] = {
            "X-App-Version": self.app_version,
            "User-Agent": f"Granola/{self.app_version} Electron/{self.electron_version} Chrome/{self.chrome_version} Node/{self.node_version} ({self.client_platform} {self.os_version}; {self.os_build})".strip(),
            "X-Client-Type": self.client_type,
            "X-Client-Platform": self.client_platform,
            "X-Client-Architecture": self.client_architecture,
            "X-Client-Id": f"granola-{self.client_type}-{self.app_version}",
            **self.client_headers,
        }
        if self._token:
            headers["Authorization"] = f"Bearer {self._token}"

        is_json_body = isinstance(body_data, (dict, list))
        if is_json_body:
            headers["Content-Type"] = "application/json"
            headers["Accept"] = "application/json" # Usually expect JSON back for JSON posts
        elif isinstance(body_data, str): # Assuming raw text body
            headers["Content-Type"] = "text/plain" # Or other appropriate type
            # Accept might still be JSON or text depending on endpoint
        else: # No body
             headers["Accept"] = "application/json" # Default accept for GET or bodiless POST

        last_error: Optional[Exception] = None

        for attempt in range(self.retries + 1):
            try:
                logger.debug(f"Request: {method} {url}, Attempt: {attempt + 1}/{self.retries + 1}")
                if body_data: logger.debug(f"Request Body: {json.dumps(body_data) if is_json_body else str(body_data)[:200]}")

                res = await self._client.request(
                    method, url,
                    content=json.dumps(body_data) if is_json_body else body_data, # Use content for bytes/str, json for dict/list
                    headers=headers,
                )

                logger.debug(f"Response: {res.status_code}, Headers: {res.headers}")

                if res.status_code == 429:
                    retry_after = res.headers.get("Retry-After")
                    delay = self._get_backoff_delay(attempt, retry_after)
                    err_text = await res.atext()
                    last_error = GranolaRateLimitError(
                        f"Rate limited. Retry after {retry_after or f'{delay}s'}.",
                        status_code=res.status_code, response_text=err_text,
                        retry_after=int(delay) if retry_after else None
                    )
                    if attempt < self.retries:
                        logger.warning(f"Rate limited. Retrying in {delay:.2f} seconds...")
                        await asyncio.sleep(delay); continue
                    raise last_error

                res.raise_for_status()
                return res

            except httpx.TimeoutException as e:
                last_error = GranolaTimeoutError(f"Request to {url} timed out after {self.timeout_ms}ms.")
                logger.warning(f"Timeout on attempt {attempt + 1} for {url}. Error: {e}")
                if attempt < self.retries: await asyncio.sleep(self._get_backoff_delay(attempt)); continue
                raise last_error
            except httpx.HTTPStatusError as e:
                err_text = e.response.text
                logger.error(f"HTTP Error: {e.response.status_code} for {url}. Response: {err_text}. Error: {e}")
                if e.response.status_code >= 500 and attempt < self.retries:
                    delay = self._get_backoff_delay(attempt, e.response.headers.get("Retry-After"))
                    logger.warning(f"Server error {e.response.status_code}. Retrying in {delay:.2f}s...")
                    await asyncio.sleep(delay); continue
                last_error = GranolaAPIError(
                    f"HTTP {e.response.status_code}: {err_text}",
                    status_code=e.response.status_code, response_text=err_text
                )
                raise last_error
            except httpx.RequestError as e:
                last_error = GranolaAPIError(f"Request failed: {e}")
                logger.warning(f"Request error on attempt {attempt + 1} for {url}. Error: {e}")
                if attempt < self.retries: await asyncio.sleep(self._get_backoff_delay(attempt)); continue
                raise last_error
            except Exception as e:
                last_error = GranolaAPIError(f"An unexpected error occurred: {e}")
                logger.error(f"Unexpected error: {e}", exc_info=True); raise last_error

        assert last_error is not None
        raise last_error

    async def _request_model(
        self, method: str, path: str,
        response_model: Type[T],
        payload_model: Optional[P] = None, # Pydantic model for request body
        payload_dict: Optional[Dict[str, Any]] = None # Raw dict for request body
    ) -> T:
        body_to_send: Optional[Dict[str, Any]] = None
        if payload_model:
            body_to_send = payload_model.model_dump(by_alias=True, exclude_none=True)
        elif payload_dict:
            body_to_send = payload_dict

        response = await self._request_raw(method, path, body_data=body_to_send)

        try:
            # Ensure we read the content before trying to parse, especially if streamed.
            # For httpx, .json() handles this.
            # We use model_validate_json for Pydantic.
            response_content = await response.aread() # Read content bytes
            if not response_content: # Handle empty response if model allows (e.g. Optional fields)
                # This case is tricky. If response_model expects data, empty content is an error.
                # If response_model can be valid with all defaults / Optionals, it might pass.
                # Pydantic will raise ValidationError if required fields are missing.
                logger.warning(f"Empty response content received for {method} {path}")

            return response_model.model_validate_json(response_content)
        except ValidationError as e:
            err_text = response_content.decode(response.encoding or 'utf-8', errors='replace') if response_content else "(empty response)"
            logger.error(f"Pydantic validation error for {method} {path}. Errors: {e.errors()}. Response: {err_text[:500]}")
            raise GranolaValidationError(str(e), validation_errors=e.errors(), response_text=err_text) from e
        except json.JSONDecodeError as e: # If response is not even JSON
            err_text = response_content.decode(response.encoding or 'utf-8', errors='replace') if response_content else "(empty response)"
            logger.error(f"JSON decode error for {method} {path}. Error: {e}. Response: {err_text[:500]}")
            raise GranolaAPIError(f"Failed to decode JSON response: {e.msg}", response_text=err_text)

    async def _request_void(
        self, method: str, path: str,
        payload_model: Optional[P] = None,
        payload_dict: Optional[Dict[str, Any]] = None
    ) -> None:
        body_to_send: Optional[Dict[str, Any]] = None
        if payload_model:
            body_to_send = payload_model.model_dump(by_alias=True, exclude_none=True)
        elif payload_dict:
            body_to_send = payload_dict

        response = await self._request_raw(method, path, body_data=body_to_send)
        # For void methods, we typically expect 204 No Content or 200/201 with empty/ignorable body.
        # raise_for_status() in _request_raw already checks for >=400 errors.
        # We might want to consume the response body to free resources, though httpx might do this.
        await response.aread()
        return None

    async def get_text(self, path: str) -> str:
        response = await self._request_raw("GET", path)
        return await response.aread().then(lambda b: b.decode(response.encoding or 'utf-8'))


    async def close(self) -> None:
        await self._client.aclose()

    async def __aenter__(self) -> "HttpClient":
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self.close()



from dotenv import load_dotenv

import pytest
import os

from granola_client import GranolaClient

load_dotenv()

@pytest.mark.asyncio
async def test_get_documents_and_metadata():
    token = os.getenv("GRANOLA_TOKEN", None)
    async with GranolaClient(token=token) as client:
        docs_response = await client.get_documents()
        assert docs_response.docs, "Should return at least one document"
        first_doc = docs_response.docs[0]
        assert hasattr(first_doc, "document_id")

        # Fetch metadata
        meta = await client.get_document_metadata(first_doc.document_id)
        # The server response does not echo document_id, so we test for creator presence instead
        assert meta.creator is not None

        # Fetch transcript
        transcript = await client.get_document_transcript(first_doc.document_id)
        assert isinstance(transcript, list)

@pytest.mark.asyncio
async def test_get_people():
    token = os.getenv("GRANOLA_TOKEN", None)
    async with GranolaClient(token=token) as client:
        resp = await client.get_people()
        assert isinstance(resp, list)
        assert len(resp) > 0
        assert hasattr(resp[0], "id")
        assert hasattr(resp[0], "name")

@pytest.mark.asyncio
async def test_get_feature_flags():
    token = os.getenv("GRANOLA_TOKEN", None)
    async with GranolaClient(token=token) as client:
        flags = await client.get_feature_flags()
        assert isinstance(flags, list)
        assert len(flags) > 0
        assert hasattr(flags[0], "feature")
        assert hasattr(flags[0], "value")

@pytest.mark.asyncio
async def test_get_panel_templates():
    token = os.getenv("GRANOLA_TOKEN", None)
    async with GranolaClient(token=token) as client:
        templates = await client.get_panel_templates()
        assert isinstance(templates, list)
        assert hasattr(templates[0], "id")
        assert hasattr(templates[0], "title")

@pytest.mark.asyncio
async def test_get_subscriptions():
    token = os.getenv("GRANOLA_TOKEN", None)
    async with GranolaClient(token=token) as client:
        subscriptions = await client.get_subscriptions()
        assert hasattr(subscriptions, "active_plan_id")
        assert hasattr(subscriptions, "subscription_plans")
        assert isinstance(subscriptions.subscription_plans, list)



import asyncio
from granola_client import GranolaClient

# Make your main function async
async def main():
    client = GranolaClient() # Assuming default init is okay, or provide token/opts
    print("\nAttempting to get documents directly...")
    try:
        # Await the async method
        # Get documents directly without workspace filters
        documents_response = await client.get_documents()
        print("\nDocuments Response (Pydantic Model):")
        print(documents_response)

        if documents_response and documents_response.docs:
            print(f"\nFound {len(documents_response.docs)} documents:")
            for doc in documents_response.docs:
                print(f"  - ID: {doc.document_id}, Title: {doc.title}")
                transcript = await client.get_transcript(doc.document_id)
                print(f"    Transcript: {transcript}")
        else:
            print("No documents found or response was empty.")

    except Exception as e:
        print(f"An error occurred: {e}")
    finally:
        await client.close()
        print("\nClient session closed.")


if __name__ == "__main__":
    # Run the async main function using asyncio.run()
    asyncio.run(main())



from typing import Callable, Awaitable, TypeVar, AsyncGenerator, Generic, Optional, List
from pydantic import BaseModel, Field, ConfigDict

T = TypeVar("T", bound=BaseModel)

class PaginatedResponse(BaseModel, Generic[T]):
    items: List[T]
    next_cursor: Optional[str] = Field(None, alias="next_cursor")

    model_config = ConfigDict(populate_by_name=True)

# FetchPageFunc is already generic with respect to T due to its definition
FetchPageFunc = Callable[[Optional[str]], Awaitable[PaginatedResponse[T]]]

# When using FetchPageFunc as a type hint for an argument,
# the type variable T is bound by the context of the function 'paginate'.
# The 'paginate' function itself is generic in T.
async def paginate(fetch_page: FetchPageFunc) -> AsyncGenerator[T, None]: # Corrected line
    """
    Async pagination iterator helper.
    :param fetch_page: An async function that takes an optional cursor string
                       and returns a PaginatedResponse object containing items of type T.
    """
    cursor: Optional[str] = None
    while True:
        # The type of 'response' will be PaginatedResponse[T]
        # because 'fetch_page' is expected to return that.
        response = await fetch_page(cursor)
        for item in response.items: # 'item' will be of type T
            yield item

        if response.next_cursor:
            cursor = response.next_cursor
        else:
            break



from typing import Any


class GranolaAPIError(Exception):
    """Base exception for Granola API errors."""
    def __init__(self, message: str, status_code: int | None = None, response_text: str | None = None):
        super().__init__(message)
        self.status_code = status_code
        self.response_text = response_text

    def __str__(self):
        if self.status_code:
            return f"HTTP {self.status_code}: {super().__str__()} - {self.response_text or ''}"
        return super().__str__()

class GranolaAuthError(GranolaAPIError):
    """Exception for authentication-related errors."""
    def __init__(self, message: str = "Authentication failed or token could not be retrieved."):
        super().__init__(message)

class GranolaRateLimitError(GranolaAPIError):
    """Exception for rate limit errors (HTTP 429)."""
    def __init__(self, message: str, status_code: int = 429, response_text: str | None = None, retry_after: int | None = None):
        super().__init__(message, status_code, response_text)
        self.retry_after = retry_after # seconds

class GranolaTimeoutError(GranolaAPIError):
    """Exception for request timeouts."""
    def __init__(self, message: str = "Request timed out."):
        super().__init__(message)

class GranolaValidationError(GranolaAPIError):
    """Exception for Pydantic validation errors on API responses."""
    def __init__(self, message: str, validation_errors: Any = None, response_text: str | None = None):
        super().__init__(f"Response validation failed: {message}")
        self.validation_errors = validation_errors
        self.response_text = response_text

    def __str__(self):
        return f"{super().__str__()} (Details: {self.validation_errors})"



from typing import List, Dict, Any, Optional

from pydantic import BaseModel, Field, HttpUrl, ConfigDict, computed_field


# Client Options
class HttpOpts(BaseModel):
    timeout: Optional[int] = 10000  # milliseconds
    retries: Optional[int] = 3
    app_version: Optional[str] = Field(default="6.4.0", alias="appVersion")
    client_type: Optional[str] = Field(
        default="electron", alias="clientType"
    )  # Consider "python-httpx"
    client_platform: Optional[str] = Field(
        default=None, alias="clientPlatform"
    )  # Will be auto-detected
    client_architecture: Optional[str] = Field(
        default=None, alias="clientArchitecture"
    )  # Will be auto-detected
    electron_version: Optional[str] = Field(default="33.4.5", alias="electronVersion")
    chrome_version: Optional[str] = Field(
        default="130.0.6723.191", alias="chromeVersion"
    )
    node_version: Optional[str] = Field(default="20.18.3", alias="nodeVersion")
    os_version: Optional[str] = Field(
        default=None, alias="osVersion"
    )  # Will be auto-detected
    os_build: Optional[str] = Field(default="", alias="osBuild")
    client_headers: Optional[Dict[str, str]] = Field(
        default_factory=dict, alias="clientHeaders"
    )

    model_config = ConfigDict(populate_by_name=True, extra="ignore")


class ClientOpts(HttpOpts):
    base_url: Optional[HttpUrl] = Field(
        default="https://api.granola.ai", alias="baseUrl"
    )


# API Response/Payload Models (examples, expand as needed based on actual API schema)


class Document(BaseModel):
    document_id: str = Field(..., alias="id")  # Assuming API might use camelCase
    title: str | None = None
    workspace_id: str | None = Field(..., alias="workspace_id")
    created_at: str = Field(..., alias="created_at")
    updated_at: str = Field(..., alias="updated_at")

    user_id: str | None = Field(default=None, alias="user_id")
    notes_markdown: str | None = Field(default=None, alias="notes_markdown")
    notes_plain: str | None = Field(default=None, alias="notes_plain")
    overview: str | None = Field(default=None)

    # Support for shared folder notes via last_viewed_panel
    last_viewed_panel: Optional[Dict[str, Any]] = Field(
        None, alias="last_viewed_panel", repr=False
    )

    @computed_field
    @property
    def notes(self) -> str:
        """Returns the meeting notes converted from ProseMirror JSON to Markdown."""
        if not self.last_viewed_panel:
            return ""

        content = self.last_viewed_panel.get("content")
        if not content:
            return ""

        # Content should be a dict/object, not a string
        if isinstance(content, str):
            return ""

        try:
            from .utils import convert_prosemirror_to_markdown

            return convert_prosemirror_to_markdown(content)
        except Exception:
            # If conversion fails, return empty string rather than crashing
            return ""

    model_config = ConfigDict(populate_by_name=True, extra="ignore")


class DocumentsResponse(BaseModel):
    docs: List[Document]
    next_cursor: Optional[str] = Field(None, alias="next_cursor")

    model_config = ConfigDict(populate_by_name=True, extra="ignore")


class Person(BaseModel):
    id: Optional[str] = None
    name: Optional[str] = None
    email: Optional[str] = None
    details: Optional[Dict[str, Any]] = None



# The API returns a list, not an object w/ a `people` key!
PeopleResponse = List[Person]


# The API returns a list of features, not `{"flags": ...}`
class FeatureFlag(BaseModel):
    feature: str
    value: Any = None
    user_id: Optional[str] = None


FeatureFlagsResponse = List[FeatureFlag]


class NotionWorkspace(BaseModel):
    id: str
    name: str


class NotionIntegrationResponse(BaseModel):
    canIntegrate: bool
    isConnected: bool
    authUrl: Optional[str] = None
    integrations: Optional[dict] = None
    # If there are additional fields, add as needed.


class SubscriptionPlan(BaseModel):
    id: str
    type: str
    display_name: str
    price: dict
    currency_iso: str
    requires_workspace: bool
    requires_payment: bool
    privacy_mode: str
    is_team_upsell_target: bool
    features: list
    display_order: int
    live: bool


class SubscriptionsResponse(BaseModel):
    active_plan_id: Optional[str] = None
    subscription_plans: Optional[list] = None


class DocumentMetadataCreator(
    BaseModel
):  # Simplified from Person for this context if structure differs
    id: Optional[str] = None
    name: str
    email: Optional[str] = None  # Example, adjust to actual API


class DocumentMetadata(BaseModel):
    document_id: Optional[str] = Field(None, alias="document_id")
    creator: DocumentMetadataCreator  # Could be Person if it matches
    attendees: Optional[List[Person]] = None  # Assuming attendees are Person objects
    # ... other fields

    model_config = ConfigDict(populate_by_name=True)


class TranscriptSegment(BaseModel):
    start_timestamp: str = Field(..., alias="startTimestamp")
    end_timestamp: str = Field(..., alias="endTimestamp")
    text: str

    model_config = ConfigDict(populate_by_name=True)


class PanelTemplate(BaseModel):
    id: str
    title: str
    # The server may not return 'templateType', so make it optional and match what the API uses (e.g., category)
    template_type: Optional[str] = Field(
        None, alias="templateType"
    )  # If API uses category, also add:
    category: Optional[str] = None

    model_config = ConfigDict(populate_by_name=True, extra="allow")


# Payload types for update methods
class UpdateDocumentPayload(BaseModel):
    document_id: str = Field(..., alias="documentId")
    title: Optional[str] = None
    notes: Optional[Dict[str, Any]] = None
    overview: Optional[str] = None
    notes_plain: Optional[str] = Field(None, alias="notesPlain")
    notes_markdown: Optional[str] = Field(None, alias="notesMarkdown")
    # Pydantic models are strict by default; unknown fields cause errors unless extra='allow'

    model_config = ConfigDict(populate_by_name=True, extra="allow")


class UpdateDocumentPanelPayload(BaseModel):
    document_id: str = Field(..., alias="documentId")
    panel_id: str = Field(..., alias="panelId")
    content: Dict[str, Any]

    model_config = ConfigDict(populate_by_name=True)


# Filters for get_documents and list_all_documents
class GetDocumentsFilters(BaseModel):
    workspace_id: Optional[str] = Field(None, alias="workspaceId")
    cursor: Optional[str] = None
    limit: Optional[int] = None

    model_config = ConfigDict(populate_by_name=True, extra="allow")


# Document Set (lightweight document index)
class DocumentSetResponse(BaseModel):
    documents: Dict[str, Dict[str, Any]]  # {doc_id: {updated_at, owner}}
    model_config = ConfigDict(populate_by_name=True, extra="allow")


# Document List (Shared Folder) Models
class DocumentListIcon(BaseModel):
    type: str
    color: str
    value: str


class DocumentListMember(BaseModel):
    user_id: str = Field(..., alias="user_id")
    name: str
    email: str
    avatar: Optional[str] = None
    role: str
    created_at: str = Field(..., alias="created_at")
    model_config = ConfigDict(populate_by_name=True)


class SlackChannel(BaseModel):
    id: str
    name: str


class DocumentList(BaseModel):
    id: str
    title: str
    description: Optional[str] = None
    icon: Optional[DocumentListIcon] = None
    visibility: str
    workspace_id: str = Field(..., alias="workspace_id")
    is_favourited: bool = Field(..., alias="is_favourited")
    user_role: Optional[str] = Field(None, alias="user_role")
    members: List[DocumentListMember]
    document_ids: List[str] = Field(..., alias="document_ids")
    slack_channel: Optional[SlackChannel] = Field(None, alias="slack_channel")
    is_shared: bool = Field(..., alias="is_shared")
    sharing_link_visibility: str = Field(..., alias="sharing_link_visibility")
    created_at: str = Field(..., alias="created_at")
    updated_at: str = Field(..., alias="updated_at")
    model_config = ConfigDict(populate_by_name=True, extra="allow")


class DocumentListsResponse(BaseModel):
    lists: Dict[str, DocumentList]
    model_config = ConfigDict(populate_by_name=True)


# Enhanced filters for get-documents with shared folder support
class EnhancedGetDocumentsFilters(GetDocumentsFilters):
    include_last_viewed_panel: Optional[bool] = Field(
        None, alias="include_last_viewed_panel"
    )
    include_shared: Optional[bool] = Field(None, alias="include_shared")
    include_folders: Optional[bool] = Field(None, alias="include_folders")
    expand_folders: Optional[bool] = Field(None, alias="expand_folders")
    show_organization: Optional[bool] = Field(None, alias="show_organization")
    document_ids: Optional[List[str]] = Field(None, alias="document_ids")
    model_config = ConfigDict(populate_by_name=True, extra="allow")



"""
Utility functions for granola-client.
"""

from typing import Dict, Any


def convert_prosemirror_to_markdown(prosemirror_json: Dict[str, Any]) -> str:
    """
    Convert ProseMirror JSON to Markdown format.
    This is based on reverse-engineering the Granola format.

    Args:
        prosemirror_json: The ProseMirror document JSON structure

    Returns:
        str: Markdown representation of the document
    """

    def process_node(node: Dict[str, Any], depth: int = 0) -> str:
        node_type = node.get("type", "")
        content = node.get("content", [])
        text = node.get("text", "")
        marks = node.get("marks", [])
        attrs = node.get("attrs", {})

        # Handle text nodes with marks
        if node_type == "text" or text:
            formatted_text = text
            for mark in marks:
                mark_type = mark.get("type", "")
                if mark_type == "bold":
                    formatted_text = f"**{formatted_text}**"
                elif mark_type == "italic":
                    formatted_text = f"*{formatted_text}*"
                elif mark_type == "code":
                    formatted_text = f"`{formatted_text}`"
                elif mark_type == "strike":
                    formatted_text = f"~~{formatted_text}~~"
                elif mark_type == "link":
                    href = mark.get("attrs", {}).get("href", "")
                    formatted_text = f"[{formatted_text}]({href})"
            return formatted_text

        # Handle different node types
        result = []

        if node_type == "doc":
            # Process all child nodes
            for child in content:
                result.append(process_node(child, depth))

        elif node_type == "paragraph":
            # Join text content and add double newline
            para_content = "".join(process_node(child, depth) for child in content)
            if para_content:  # Only add non-empty paragraphs
                result.append(para_content + "\n")

        elif node_type == "heading":
            level = attrs.get("level", 1)
            heading_text = "".join(process_node(child, depth) for child in content)
            result.append("#" * level + " " + heading_text + "\n")

        elif node_type == "bulletList":
            for item in content:
                result.append(process_node(item, depth))

        elif node_type == "orderedList":
            for i, item in enumerate(content, 1):
                # Pass the order number in attrs for listItem to use
                item_attrs = item.get("attrs", {})
                item_attrs["order"] = i
                item["attrs"] = item_attrs
                result.append(process_node(item, depth))

        elif node_type == "listItem":
            # Process list item content
            item_content = []
            for child in content:
                item_content.append(process_node(child, depth + 1))

            # Format based on parent list type (detect by order attr)
            order = attrs.get("order")
            if order:
                prefix = f"{order}. "
            else:
                prefix = "- "

            # Add proper indentation
            indent = "  " * depth
            formatted_content = "".join(item_content).strip()
            result.append(indent + prefix + formatted_content + "\n")

        elif node_type == "codeBlock":
            language = attrs.get("language", "")
            code_content = "".join(process_node(child, depth) for child in content)
            result.append(f"```{language}\n{code_content}\n```\n")

        elif node_type == "blockquote":
            quote_content = "".join(process_node(child, depth) for child in content)
            # Add > prefix to each line
            quoted_lines = ["> " + line for line in quote_content.strip().split("\n")]
            result.append("\n".join(quoted_lines) + "\n")

        elif node_type == "horizontalRule":
            result.append("---\n")

        elif node_type == "hardBreak":
            result.append("  \n")  # Markdown hard break

        else:
            # For unknown types, just process children
            for child in content:
                result.append(process_node(child, depth))

        return "".join(result)

    # Start processing from the root
    return process_node(prosemirror_json).strip()



import json
import platform
from pathlib import Path
from typing import Optional, AsyncGenerator, Tuple, cast, List
import logging

from pydantic import HttpUrl, BaseModel, ValidationError

from .http_client import HttpClient
from .types import (
    ClientOpts,  # These are now Pydantic models
    Document,
    DocumentsResponse,
    DocumentMetadata,
    TranscriptSegment,
    PanelTemplate,
    NotionIntegrationResponse,
    SubscriptionsResponse,
    UpdateDocumentPayload,
    UpdateDocumentPanelPayload,
    GetDocumentsFilters,
    Person,
    FeatureFlag,
)
from .pagination import paginate, PaginatedResponse
from .errors import GranolaAuthError, GranolaValidationError

logger = logging.getLogger(__name__)


class GranolaClient:
    http: HttpClient

    def __init__(self, token: Optional[str] = None, opts: Optional[ClientOpts] = None):
        # opts can be a dict or ClientOpts instance. Pydantic handles validation.
        # If opts is None, ClientOpts() provides defaults.
        client_options = ClientOpts.model_validate(opts if opts is not None else {})

        self.http = HttpClient(
            token=token,
            base_url=cast(
                HttpUrl, client_options.base_url
            ),  # base_url is HttpUrl from ClientOpts
            opts=client_options,  # Pass the validated Pydantic model
        )

        if not token:
            if platform.system() == "Darwin":
                self.http.set_token_provider(self._provide_auth_token_macos)
            else:
                logger.warning(
                    "Automatic token retrieval is macOS only. Provide token manually."
                )

    async def _provide_auth_token_macos(self) -> str:
        tokens = await GranolaClient.get_auth_tokens()
        return tokens[0]

    @staticmethod
    async def get_auth_tokens() -> Tuple[str, str]:
        if platform.system() != "Darwin":
            raise GranolaAuthError("Automatic token extraction is macOS only.")
        try:
            supabase_file_path = (
                Path.home() / "Library/Application Support/Granola/supabase.json"
            )
            if not supabase_file_path.exists():
                raise GranolaAuthError(f"Token file not found: {supabase_file_path}")

            data = json.loads(supabase_file_path.read_text(encoding="utf-8"))
            cognito_tokens = json.loads(data["cognito_tokens"])
            access, refresh = (
                cognito_tokens["access_token"],
                cognito_tokens["refresh_token"],
            )
            if not access or not refresh:
                raise GranolaAuthError("Access or refresh token missing.")
            return access, refresh
        except (FileNotFoundError, KeyError, json.JSONDecodeError) as e:
            logger.error(f"Failed to get auth tokens: {e}", exc_info=True)
            raise GranolaAuthError(f"Failed to extract auth tokens: {e}") from e

    async def get_documents(
        self, filters: Optional[GetDocumentsFilters] = None
    ) -> DocumentsResponse:
        payload = (
            filters.model_dump(by_alias=True, exclude_none=True) if filters else {}
        )
        return await self.http._request_model(
            "POST",
            "/v2/get-documents",
            response_model=DocumentsResponse,
            payload_dict=payload,
        )

    async def list_all_documents(
        self, filters: Optional[GetDocumentsFilters] = None
    ) -> AsyncGenerator[Document, None]:
        initial_filters_dict = (
            filters.model_dump(by_alias=True, exclude_none=True, exclude={"cursor"})
            if filters
            else {}
        )

        async def fetch_page_func(
            cursor: Optional[str] = None,
        ) -> PaginatedResponse[Document]:
            current_filters_payload = {**initial_filters_dict}
            if cursor:
                current_filters_payload["cursor"] = cursor

            # get_documents expects GetDocumentsFilters model or dict.
            # Since we constructed a dict, we pass it directly.
            # The response from get_documents is DocumentsResponse.
            # We need to adapt it to PaginatedResponse[Document].
            docs_response = await self.http._request_model(
                "POST",
                "/v2/get-documents",
                response_model=DocumentsResponse,  # This has docs and next_cursor
                payload_dict=current_filters_payload,
            )
            return PaginatedResponse[Document](
                items=docs_response.docs, next_cursor=docs_response.next_cursor
            )

        async for doc in paginate(fetch_page_func):
            yield doc

    async def get_document_metadata(self, document_id: str) -> DocumentMetadata:
        return await self.http._request_model(
            "POST",
            "/v1/get-document-metadata",
            response_model=DocumentMetadata,
            payload_dict={"document_id": document_id},
        )

    async def get_document_transcript(
        self, document_id: str
    ) -> List[TranscriptSegment]:
        # Assuming the API returns a direct list of segments, not nested in a model.
        # If it's nested, e.g. {"segments": [...]}, then define a wrapper Pydantic model.
        # For now, let's assume _request_raw and parse manually if it's a direct list.
        # This is a common pattern Pydantic v2 handles with RootModel or TypeAdapter.
        # A simpler way if http_client always returns a model:
        class TranscriptResponse(
            BaseModel
        ):  # Temporary wrapper if API returns {"transcript": [...]} or similar
            items: List[
                TranscriptSegment
            ]  # Or RootModel[List[TranscriptSegment]] if it's just a list

        # If API directly returns a JSON list `[...]`
        # Use TypeAdapter for direct list parsing
        from pydantic import TypeAdapter

        ta = TypeAdapter(List[TranscriptSegment])

        response = await self.http._request_raw(
            "POST",
            "/v1/get-document-transcript",
            body_data={"document_id": document_id},
        )
        response_content = await response.aread()
        try:
            return ta.validate_json(response_content)
        except ValidationError as e:
            err_text = response_content.decode(
                response.encoding or "utf-8", errors="replace"
            )
            raise GranolaValidationError(
                str(e), validation_errors=e.errors(), response_text=err_text
            )

    async def update_document(self, payload: UpdateDocumentPayload) -> None:
        await self.http._request_void(
            "POST", "/v1/update-document", payload_model=payload
        )

    async def update_document_panel(self, payload: UpdateDocumentPanelPayload) -> None:
        await self.http._request_void(
            "POST", "/v1/update-document-panel", payload_model=payload
        )

    async def get_panel_templates(self) -> List[PanelTemplate]:
        # The API returns a list.
        from pydantic import TypeAdapter

        ta = TypeAdapter(List[PanelTemplate])
        response = await self.http._request_raw(
            "POST", "/v1/get-panel-templates", body_data={}
        )
        response_content = await response.aread()
        try:
            return ta.validate_json(response_content)
        except ValidationError as e:
            err_text = response_content.decode(
                response.encoding or "utf-8", errors="replace"
            )
            raise GranolaValidationError(
                str(e), validation_errors=e.errors(), response_text=err_text
            )

    async def get_people(self) -> List["Person"]:
        # API returns a list, not an object with 'people' key.
        from pydantic import TypeAdapter

        ta = TypeAdapter(List[Person])
        response = await self.http._request_raw("POST", "/v1/get-people", body_data={})
        response_content = await response.aread()
        try:
            return ta.validate_json(response_content)
        except ValidationError as e:
            err_text = response_content.decode(
                response.encoding or "utf-8", errors="replace"
            )
            raise GranolaValidationError(
                str(e), validation_errors=e.errors(), response_text=err_text
            )

    async def get_feature_flags(self) -> List["FeatureFlag"]:
        from pydantic import TypeAdapter

        ta = TypeAdapter(List[FeatureFlag])
        response = await self.http._request_raw(
            "POST", "/v1/get-feature-flags", body_data={}
        )
        response_content = await response.aread()
        try:
            return ta.validate_json(response_content)
        except ValidationError as e:
            err_text = response_content.decode(
                response.encoding or "utf-8", errors="replace"
            )
            raise GranolaValidationError(
                str(e), validation_errors=e.errors(), response_text=err_text
            )

    async def get_notion_integration(self) -> NotionIntegrationResponse:
        return await self.http._request_model(
            "POST",
            "/v1/get-notion-integration",
            response_model=NotionIntegrationResponse,
            payload_dict={},
        )

    async def get_subscriptions(self) -> SubscriptionsResponse:
        return await self.http._request_model(
            "POST",
            "/v1/get-subscriptions",
            response_model=SubscriptionsResponse,
            payload_dict={},
        )

    async def refresh_google_events(self) -> None:
        await self.http._request_void(
            "POST", "/v1/refresh-google-events", payload_dict={}
        )

    async def check_for_update(self) -> str:
        platform_path = "latest-mac.yml"  # Default
        system = platform.system()
        if system == "Windows":
            platform_path = "latest.yml"  # Common for Windows
        elif system == "Linux":
            platform_path = "latest-linux.yml"

        return await self.http.get_text(f"/v1/check-for-update/{platform_path}")

    async def get_document_set(self) -> "DocumentSetResponse":
        """Get a lightweight index of all documents user has access to."""
        from .types import DocumentSetResponse

        return await self.http._request_model(
            "POST",
            "/v1/get-document-set",
            response_model=DocumentSetResponse,
            payload_dict={},
        )

    async def get_document_lists(self) -> "DocumentListsResponse":
        """Get metadata for all shared folders/document lists."""
        from .types import DocumentListsResponse

        return await self.http._request_model(
            "POST",
            "/v1/get-document-lists-metadata",
            response_model=DocumentListsResponse,
            payload_dict={
                "include_document_ids": True,
                "include_only_joined_lists": False,
            },
        )

    async def get_documents_enhanced(
        self, filters: "EnhancedGetDocumentsFilters"
    ) -> DocumentsResponse:
        """Enhanced version of get_documents with support for shared folder features."""
        payload = filters.model_dump(by_alias=True, exclude_none=True)
        return await self.http._request_model(
            "POST",
            "/v2/get-documents",
            response_model=DocumentsResponse,
            payload_dict=payload,
        )

    async def get_documents_by_folder_id(self, folder_id: str) -> List[Document]:
        """Get all documents from a specific shared folder/document list by ID."""
        from .types import EnhancedGetDocumentsFilters

        # Get folder metadata to find document IDs
        folders_response = await self.get_document_lists()

        # Find the target folder
        target_folder = folders_response.lists.get(folder_id)
        if not target_folder:
            raise ValueError(f"Folder with ID '{folder_id}' not found")

        # Get document IDs from the folder
        document_ids = target_folder.document_ids
        if not document_ids:
            return []

        # Fetch documents with enhanced filters to get notes
        filters = EnhancedGetDocumentsFilters(
            document_ids=document_ids,
            include_last_viewed_panel=True,
            include_shared=True,
            include_folders=True,
            expand_folders=True,
            show_organization=True,
        )

        response = await self.get_documents_enhanced(filters)
        return response.docs

    async def get_documents_by_folder_name(
        self, folder_name: str, case_sensitive: bool = False
    ) -> List[Document]:
        """Get all documents from a specific shared folder/document list by name."""
        # Get folder metadata
        folders_response = await self.get_document_lists()

        # Find folder(s) by name
        matching_folders = []
        for folder_id, folder_data in folders_response.lists.items():
            if case_sensitive:
                if folder_data.title == folder_name:
                    matching_folders.append((folder_id, folder_data))
            else:
                if folder_data.title.lower() == folder_name.lower():
                    matching_folders.append((folder_id, folder_data))

        if not matching_folders:
            raise ValueError(f"No folder found with name '{folder_name}'")
        elif len(matching_folders) > 1:
            folder_titles = [f.title for _, f in matching_folders]
            raise ValueError(
                f"Multiple folders found with name '{folder_name}': {folder_titles}. "
                "Use get_documents_by_folder_id() with a specific folder ID instead."
            )

        # Use the single matching folder
        folder_id, _ = matching_folders[0]
        return await self.get_documents_by_folder_id(folder_id)

    async def close(self) -> None:
        await self.http.close()

    async def __aenter__(self) -> "GranolaClient":
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self.close()
